
#include <stdio.h>
#include <Arduino.h>
#include <EEPROM.h>
#include <SPI.h>
#include <avr/pgmspace.h>

#include "FT_Platform.h"
#include "sdcard.h"
#ifdef MSVC_PLATFORM
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <windows.h>
#include <io.h>
#include <direct.h>
#include <time.h>
#endif

/* Global variables for display resolution to support various display panels */
/* Default is WQVGA - 480x272 */
ft_int16_t FT_DispWidth = 480;
ft_int16_t FT_DispHeight = 272;
ft_int16_t FT_DispHCycle =  548;
ft_int16_t FT_DispHOffset = 43;
ft_int16_t FT_DispHSync0 = 0;
ft_int16_t FT_DispHSync1 = 41;
ft_int16_t FT_DispVCycle = 292;
ft_int16_t FT_DispVOffset = 12;
ft_int16_t FT_DispVSync0 = 0;
ft_int16_t FT_DispVSync1 = 10;
ft_uint8_t FT_DispPCLK = 5;
ft_char8_t FT_DispSwizzle = 0;
ft_char8_t FT_DispPCLKPol = 1;

/* Global used for buffer optimization */
Ft_Gpu_Hal_Context_t host,*phost;

ft_uint32_t Ft_CmdBuffer_Index;
ft_uint32_t Ft_DlBuffer_Index;

#ifdef BUFFER_OPTIMIZATION
ft_uint8_t  Ft_DlBuffer[FT_DL_SIZE];
ft_uint8_t  Ft_CmdBuffer[FT_CMD_FIFO_SIZE];
#endif

ft_void_t Ft_App_WrCoCmd_Buffer(Ft_Gpu_Hal_Context_t *phost,ft_uint32_t cmd)
{
#ifdef  BUFFER_OPTIMIZATION
   /* Copy the command instruction into buffer */
   ft_uint32_t *pBuffcmd;
   pBuffcmd =(ft_uint32_t*)&Ft_CmdBuffer[Ft_CmdBuffer_Index];
   *pBuffcmd = cmd;
#endif
#ifdef ARDUINO_PLATFORM
   Ft_Gpu_Hal_WrCmd32(phost,cmd);
#endif

   /* Increment the command index */
   Ft_CmdBuffer_Index += FT_CMD_SIZE;  
}

ft_void_t Ft_App_WrDlCmd_Buffer(Ft_Gpu_Hal_Context_t *phost,ft_uint32_t cmd)
{
#ifdef BUFFER_OPTIMIZATION  
   /* Copy the command instruction into buffer */
   ft_uint32_t *pBuffcmd;
   pBuffcmd =(ft_uint32_t*)&Ft_DlBuffer[Ft_DlBuffer_Index];
   *pBuffcmd = cmd;
#endif

#ifdef ARDUINO_PLATFORM
   Ft_Gpu_Hal_Wr32(phost,(RAM_DL+Ft_DlBuffer_Index),cmd);
#endif
   /* Increment the command index */
   Ft_DlBuffer_Index += FT_CMD_SIZE;  
}

ft_void_t Ft_App_WrCoStr_Buffer(Ft_Gpu_Hal_Context_t *phost,const ft_char8_t *s)
{
#ifdef  BUFFER_OPTIMIZATION  
  ft_uint16_t length = 0;
  length = strlen(s) + 1;//last for the null termination
  
  strcpy(&Ft_CmdBuffer[Ft_CmdBuffer_Index],s);  

  /* increment the length and align it by 4 bytes */
  Ft_CmdBuffer_Index += ((length + 3) & ~3);  
#endif  
}

ft_void_t Ft_App_Flush_DL_Buffer(Ft_Gpu_Hal_Context_t *phost)
{
#ifdef  BUFFER_OPTIMIZATION    
   if (Ft_DlBuffer_Index> 0)
     Ft_Gpu_Hal_WrMem(phost,RAM_DL,Ft_DlBuffer,Ft_DlBuffer_Index);
#endif     
   Ft_DlBuffer_Index = 0;
   
}

ft_void_t Ft_App_Flush_Co_Buffer(Ft_Gpu_Hal_Context_t *phost)
{
#ifdef  BUFFER_OPTIMIZATION    
   if (Ft_CmdBuffer_Index > 0)
     Ft_Gpu_Hal_WrCmdBuf(phost,Ft_CmdBuffer,Ft_CmdBuffer_Index);
#endif     
   Ft_CmdBuffer_Index = 0;
}

ft_void_t Ft_BootupConfig()
{
	/* Do a power cycle for safer side */
	Ft_Gpu_Hal_Powercycle(phost,FT_TRUE);
	 Ft_Gpu_Hal_Rd16(phost,RAM_G);
	/* Set the clk to external clock */
	Ft_Gpu_HostCommand(phost,FT_GPU_EXTERNAL_OSC);  
	Ft_Gpu_Hal_Sleep(10);
	  

	/* Switch PLL output to 48MHz */
	Ft_Gpu_HostCommand(phost,FT_GPU_PLL_48M);  
	Ft_Gpu_Hal_Sleep(10);

	/* Do a core reset for safer side */
	Ft_Gpu_HostCommand(phost,FT_GPU_CORE_RESET);     

	/* Access address 0 to wake up the FT800 */
	Ft_Gpu_HostCommand(phost,FT_GPU_ACTIVE_M);  

    Ft_Gpu_Hal_Wr8(phost, REG_GPIO_DIR,0x80 | Ft_Gpu_Hal_Rd8(phost,REG_GPIO_DIR));
    Ft_Gpu_Hal_Wr8(phost, REG_GPIO,0x080 | Ft_Gpu_Hal_Rd8(phost,REG_GPIO));
	
	{
		ft_uint8_t chipid;
		//Read Register ID to check if FT800 is ready. 
		chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
		while(chipid != 0x7C)
			chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
#ifdef MSVC_PLATFORM
		printf("VC1 register ID after wake up %x\n",chipid);
#endif
	}
	/* Configuration of LCD display */
#ifdef SAMAPP_DISPLAY_QVGA	
	/* Values specific to QVGA LCD display */
	FT_DispWidth = 320;
	FT_DispHeight = 240;
	FT_DispHCycle =  408;
	FT_DispHOffset = 70;
	FT_DispHSync0 = 0;
	FT_DispHSync1 = 10;
	FT_DispVCycle = 263;
	FT_DispVOffset = 13;
	FT_DispVSync0 = 0;
	FT_DispVSync1 = 2;
	FT_DispPCLK = 8;
	FT_DispSwizzle = 2;
	FT_DispPCLKPol = 0;
#endif

    Ft_Gpu_Hal_Wr16(phost, REG_HCYCLE, FT_DispHCycle);
    Ft_Gpu_Hal_Wr16(phost, REG_HOFFSET, FT_DispHOffset);
    Ft_Gpu_Hal_Wr16(phost, REG_HSYNC0, FT_DispHSync0);
    Ft_Gpu_Hal_Wr16(phost, REG_HSYNC1, FT_DispHSync1);
    Ft_Gpu_Hal_Wr16(phost, REG_VCYCLE, FT_DispVCycle);
    Ft_Gpu_Hal_Wr16(phost, REG_VOFFSET, FT_DispVOffset);
    Ft_Gpu_Hal_Wr16(phost, REG_VSYNC0, FT_DispVSync0);
    Ft_Gpu_Hal_Wr16(phost, REG_VSYNC1, FT_DispVSync1);
    Ft_Gpu_Hal_Wr8(phost, REG_SWIZZLE, FT_DispSwizzle);
    Ft_Gpu_Hal_Wr8(phost, REG_PCLK_POL, FT_DispPCLKPol);
    Ft_Gpu_Hal_Wr8(phost, REG_PCLK,FT_DispPCLK);//after this display is visible on the LCD
    Ft_Gpu_Hal_Wr16(phost, REG_HSIZE, FT_DispWidth);
    Ft_Gpu_Hal_Wr16(phost, REG_VSIZE, FT_DispHeight);


    /* Touch configuration - configure the resistance value to 1200 - this value is specific to customer requirement and derived by experiment */
    Ft_Gpu_Hal_Wr16(phost, REG_TOUCH_RZTHRESH,1200);
    Ft_Gpu_Hal_Wr8(phost, REG_GPIO_DIR,0xff);
    Ft_Gpu_Hal_Wr8(phost, REG_GPIO,0x0ff);
}

/* Boot up for FT800 followed by graphics primitive sample cases */
/* Initial boot up DL - make the back ground green color */
const ft_uint8_t FT_DLCODE_BOOTUP[12] = 
{
  255,255,255,2,//GPU instruction CLEAR_COLOR_RGB
  7,0,0,38, //GPU instruction CLEAR
  0,0,0,0,  //GPU instruction DISPLAY
};

ft_void_t Info()
{
  ft_uint16_t dloffset = 0,z;
// Touch Screen Calibration
  Ft_CmdBuffer_Index = 0;  
  Ft_Gpu_CoCmd_Dlstart(phost); 
  Ft_App_WrCoCmd_Buffer(phost,CLEAR(1,1,1));
  Ft_App_WrCoCmd_Buffer(phost,COLOR_RGB(255,255,255));
  Ft_Gpu_CoCmd_Text(phost,FT_DispWidth/2,FT_DispHeight/2,28,OPT_CENTERX|OPT_CENTERY,"Please tap on a dot");
  Ft_Gpu_CoCmd_Calibrate(phost,0);
  Ft_App_WrCoCmd_Buffer(phost,DISPLAY());
  Ft_Gpu_CoCmd_Swap(phost);
  Ft_App_Flush_Co_Buffer(phost);
  Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
}
// >>> [int(round(1024 * math.atan(i / 256.) / math.pi)) for i in range(256)]
static const PROGMEM prog_uchar atan8[] = {
0,1,3,4,5,6,8,9,10,11,13,14,15,17,18,19,20,22,23,24,25,27,28,29,30,32,33,34,36,37,38,39,41,42,43,44,46,47,48,49,51,52,53,54,55,57,58,59,60,62,63,64,65,67,68,69,70,71,73,74,75,76,77,79,80,81,82,83,85,86,87,88,89,91,92,93,94,95,96,98,99,100,101,102,103,104,106,107,108,109,110,111,112,114,115,116,117,118,119,120,121,122,124,125,126,127,128,129,130,131,132,133,134,135,137,138,139,140,141,142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,177,178,179,180,181,182,183,184,185,186,187,188,188,189,190,191,192,193,194,195,195,196,197,198,199,200,201,201,202,203,204,205,206,206,207,208,209,210,211,211,212,213,214,215,215,216,217,218,219,219,220,221,222,222,223,224,225,225,226,227,228,228,229,230,231,231,232,233,234,234,235,236,236,237,238,239,239,240,241,241,242,243,243,244,245,245,246,247,248,248,249,250,250,251,251,252,253,253,254,255,255
};

uint16_t atan2(int16_t y, int16_t x)
{
  uint16_t a;
  uint16_t xx = 0;
  
  //hack from prashanth
  if(0x8000 == y)
  {
	y++;
  }
  if(0x8000 == x)
  {
  	x++;
  }  

  if ((x <= 0) ^ (y > 0)) {
    int16_t t; t = x; x = y; y = t;
    xx ^= 0x4000;
  }
  if (x <= 0) {
    x = -x;
  } else {
    xx ^= 0x8000;
  }
  y = abs(y);
  if (x > y) {
    int16_t t; t = x; x = y; y = t;
    xx ^= 0x3fff;
  }
  while ((x | y) & 0xff80) {
    x >>= 1;
    y >>= 1;
  }
  if (y == 0) {
    a = 0;
  } else if (x == y) {
    a = 0x2000;
  } else {
    // assert(x <= y);
    int r = ((x << 8) / y);
    // assert(0 <= r);
    // assert(r < 256);
    a = pgm_read_byte(atan8 + r) << 5;
  }
  a ^= xx;
  return a;
}

class Poly {
    int x0, y0, x1, y1;
    int x[8], y[8];
    byte n;
    void restart() {
      n = 0;
      x0 = 16 * 480;
      x1 = 0;
      y0 = 16 * 272;
      y1 = 0;
    }
    void perim() {
      for (byte i = 0; i < n; i++)
        Ft_App_WrCoCmd_Buffer(phost,VERTEX2F(x[i], y[i]));
      Ft_App_WrCoCmd_Buffer(phost,VERTEX2F(x[0], y[0]));
    }
  public:
    void begin() {
      restart();

      Ft_App_WrCoCmd_Buffer(phost,COLOR_MASK(0,0,0,0));
      Ft_App_WrCoCmd_Buffer(phost,STENCIL_OP(KEEP, INVERT));
      Ft_App_WrCoCmd_Buffer(phost,STENCIL_FUNC(ALWAYS, 255, 255));
    }
    void v(int _x, int _y) {
      x0 = min(x0, _x >> 4);
      x1 = max(x1, _x >> 4);
      y0 = min(y0, _y >> 4);
      y1 = max(y1, _y >> 4);
      x[n] = _x;
      y[n] = _y;
      n++;
    }
    void paint() {
      x0 = max(0, x0);
      y0 = max(0, y0);
      x1 = min(16 * 480, x1);
      y1 = min(16 * 272, y1);
      Ft_App_WrCoCmd_Buffer(phost,SCISSOR_XY(x0, y0));
      Ft_App_WrCoCmd_Buffer(phost,SCISSOR_SIZE(x1 - x0 + 1, y1 - y0 + 1));
      Ft_App_WrCoCmd_Buffer(phost,BEGIN(EDGE_STRIP_B));
      perim();
    }
    void outline() {
      Ft_App_WrCoCmd_Buffer(phost,BEGIN(LINE_STRIP));
      perim();
    }
};

static void read_extended(int16_t sx[5], int16_t sy[5])
{
  uint32_t sxy0, sxyA, sxyB, sxyC;
  sxy0 = Ft_Gpu_Hal_Rd32(phost,REG_CTOUCH_TOUCH0_XY);
  sxyA = Ft_Gpu_Hal_Rd32(phost,REG_CTOUCH_TOUCH1_XY);
  sxyB = Ft_Gpu_Hal_Rd32(phost,REG_CTOUCH_TOUCH2_XY);
  sxyC = Ft_Gpu_Hal_Rd32(phost,REG_CTOUCH_TOUCH3_XY);

  sx[0] = sxy0 >> 16;
  sy[0] = sxy0;
  sx[1] = sxyA >> 16;
  sy[1] = sxyA;
  sx[2] = sxyB >> 16;
  sy[2] = sxyB;
  sx[3] = sxyC >> 16;
  sy[3] = sxyC;

  sx[4] = Ft_Gpu_Hal_Rd16(phost,REG_CTOUCH_TOUCH4_X);
  sy[4] = Ft_Gpu_Hal_Rd16(phost,REG_CTOUCH_TOUCH4_Y);
  
}
int filesize = 0;
ft_uint16_t blocklen;
Reader imfile,dbfile;
char filename[1][9] = {"ftdi.jpg"};
ft_uint8_t ibuff[512];
ft_void_t Load_Jpeg()
{
  while(filesize > 0)
  { 
      blocklen = filesize>512?512:filesize;
      imfile.readsector(ibuff);
      filesize -= blocklen;
      Ft_Gpu_Hal_WrCmdBuf(phost,ibuff,blocklen);				/* copy data continuously into command memory */
      Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
  }
}

ft_void_t Loadimage2ram(ft_uint8_t Song_no,ft_uint8_t add)
{
    imfile.openfile(filename[Song_no]);
    filesize = imfile.size;
    filesize = (filesize+3) & ~3;
    Ft_Gpu_Hal_WrCmd32(phost,CMD_LOADIMAGE);
    Ft_Gpu_Hal_WrCmd32(phost,(add*24000L));
    Ft_Gpu_Hal_WrCmd32(phost,OPT_NODL);
}

#ifdef MSVC_PLATFORM
/* Main entry point */
ft_int32_t main(ft_int32_t argc,ft_char8_t *argv[])
#endif
#ifdef ARDUINO_PLATFORM
ft_void_t setup()
#endif
{
    ft_uint8_t chipid;
    static ft_uint8_t sd_present =0;
    Ft_Gpu_HalInit_t halinit;	
    halinit.TotalChannelNum = 1;
              
    Ft_Gpu_Hal_Init(&halinit);
    host.hal_config.channel_no = 0;
#ifdef MSVC_PLATFORM_SPI
    host.hal_config.spi_clockrate_khz = 12000; //in KHz
#endif
#ifdef ARDUINO_PLATFORM_SPI
    host.hal_config.spi_clockrate_khz = 4000; //in KHz
#endif
    Ft_Gpu_Hal_Open(&host);     
    phost = &host;
    Ft_BootupConfig();
  
    /*It is optional to clear the screen here*/
    Ft_Gpu_Hal_WrMem(phost, RAM_DL,(ft_uint8_t *)FT_DLCODE_BOOTUP,sizeof(FT_DLCODE_BOOTUP));
    Ft_Gpu_Hal_Wr8(phost, REG_DLSWAP,DLSWAP_FRAME);

    pinMode(FT_SDCARD_CS, OUTPUT);
    digitalWrite(FT_SDCARD_CS, HIGH);
    delay(100);
    sd_present =  SD.begin(FT_SDCARD_CS);       
    SPI.setClockDivider(SPI_CLOCK_DIV2);
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);   

    Serial.begin(9600);
    ft_int16_t xoffset,yoffset;
    xoffset = ((FT_DispWidth - 480)/2);
    yoffset = ((FT_DispHeight - 272)/2);
    Info();  
    if(sd_present)
    {
      Ft_Gpu_Hal_Wr8(phost,REG_CTOUCH_EXTENDED, REG_CTOUCH_EXTENDED);
      Loadimage2ram(0,0);
      Load_Jpeg();
      
      while(1)
      {
        Ft_App_WrCoCmd_Buffer(phost, CMD_DLSTART);
  	Ft_App_WrCoCmd_Buffer(phost, CLEAR_COLOR_RGB(0x50,0,0));
  	Ft_App_WrCoCmd_Buffer(phost, ((38UL << 24) | 7));
  	Ft_App_WrCoCmd_Buffer(phost, SAVE_CONTEXT());
  	Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(0x40,0x40,0x40));
  	Ft_App_WrCoCmd_Buffer(phost, BEGIN(BITMAPS));
  	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SOURCE(0));
  	Ft_App_WrCoCmd_Buffer(phost, BITMAP_LAYOUT(RGB565,480*2,272));
  	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SIZE(BILINEAR,BORDER,BORDER,480,272));
  	Ft_App_WrCoCmd_Buffer(phost, VERTEX2F(xoffset*16,yoffset*16));		
  	Ft_App_WrCoCmd_Buffer(phost, END());
  	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);      
  
        int16_t sx[5], sy[5];
        read_extended(sx, sy);
        
        int16_t cx = 0, cy = 0;
        byte n = 0;
        {
          for (int i = 0; i < 5; i++) {
            if (sx[i] > -10) {
              cx += sx[i];
              cy += sy[i];
              n++;
            }
          }
          cx /= n;
          cy /= n;
        }
  
        byte angles[5];
        for (int i = 0; i < 5; i++) {
          angles[i] = atan2(sx[i] - cx, sy[i] - cy) >> 8;
        }

       Poly po;
        po.begin();
        for (int a = 0; a < 256; a++) {
          for (int i = 0; i < 5; i++) {
            if (angles[i] == a && sx[i]> -10) {
              po.v(16 * sx[i], 16 * sy[i]);
            }
          }
        }
        po.paint();
        
	Ft_App_WrCoCmd_Buffer(phost,COLOR_A(255));
	Ft_App_WrCoCmd_Buffer(phost,COLOR_MASK(1,1,1,1));
	Ft_App_WrCoCmd_Buffer(phost,BLEND_FUNC(SRC_ALPHA,ONE));
	Ft_App_WrCoCmd_Buffer(phost,STENCIL_FUNC(EQUAL, 255, 255));
	Ft_App_WrCoCmd_Buffer(phost, BEGIN(RECTS));
	Ft_App_WrCoCmd_Buffer(phost,VERTEX2II(0,0,0,0));
	Ft_App_WrCoCmd_Buffer(phost,VERTEX2II(480,272,0,0));

	Ft_App_WrCoCmd_Buffer(phost,RESTORE_CONTEXT());
	Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(0x80,0x80,0xff));
	Ft_App_WrCoCmd_Buffer(phost, LINE_WIDTH(24));
	po.outline();        

        Ft_App_WrCoCmd_Buffer(phost,DISPLAY());
        Ft_Gpu_CoCmd_Swap(phost);
        Ft_Gpu_CoCmd_LoadIdentity(phost);
        Ft_App_Flush_Co_Buffer(phost);	
      }
    }
    else
    {
        Ft_Gpu_CoCmd_Dlstart(phost);    
        Ft_App_WrCoCmd_Buffer(phost,CLEAR(1,1,1));
        Ft_App_WrCoCmd_Buffer(phost,SCISSOR_XY(FT_DispWidth/12,FT_DispHeight/9));
        Ft_App_WrCoCmd_Buffer(phost,SCISSOR_SIZE(FT_DispWidth-2*(FT_DispWidth/12),FT_DispHeight-2*(FT_DispHeight/9)));
        Ft_App_WrCoCmd_Buffer(phost,CLEAR_COLOR_RGB(26,98,161));
        Ft_App_WrCoCmd_Buffer(phost,CLEAR(1,1,1));
        Ft_Gpu_CoCmd_Text(phost,FT_DispWidth/2,FT_DispHeight/2,28,OPT_CENTERX|OPT_CENTERY,"Storage Device not Found");
        Ft_App_WrCoCmd_Buffer(phost,DISPLAY());
        Ft_Gpu_CoCmd_Swap(phost);
        Ft_App_Flush_Co_Buffer(phost);
        Ft_Gpu_Hal_WaitCmdfifo_empty(phost);	
        delay(5000);   
    }
    Ft_Gpu_Hal_Close(phost);
    Ft_Gpu_Hal_DeInit();
#ifdef MSVC_PLATFORM
	return 0;
#endif
}

void loop()
{
}
/* Nothing beyond this */














